from yadt.const import (
    CACHE_FOLDER,
)
import os


class TranslationConfig:
    def __init__(
        self,
        input_file: str,
        translator,
        lang_in: str,
        lang_out: str,
        font: str | None = None,
        pages: str | None = None,
            output_dir: str | None = None,
        debug: bool = False,
        working_dir: str | None = None,
        no_dual: bool = False,
        no_mono: bool = False,
        formular_font_pattern: str | None = None,
        formular_char_pattern: str | None = None,
        qps: int = 1,
    ):
        self.input_file = input_file
        self.translator = translator
        self.font = font
        self.pages = pages
        self.page_ranges = self._parse_pages(pages) if pages else None
        self.debug = debug
        self.lang_in = lang_in
        self.lang_out = lang_out
        self.no_dual = no_dual
        self.no_mono = no_mono
        self.formular_font_pattern = formular_font_pattern
        self.formular_char_pattern = formular_char_pattern
        self.qps = qps

        if working_dir is None:
            working_dir = os.path.join(
                CACHE_FOLDER, "working", os.path.basename(input_file).split(".")[0]
            )
        self.working_dir = working_dir

        os.makedirs(working_dir, exist_ok=True)

        if output_dir is None:
            output_dir = os.path.dirname(input_file)
        self.output_dir = output_dir

        os.makedirs(output_dir, exist_ok=True)

    def get_output_file_path(self, filename):
        return os.path.join(self.output_dir, filename)

    def get_working_file_path(self, filename):
        return os.path.join(self.working_dir, filename)

    def _parse_pages(self, pages_str: str | None) -> list[tuple[int, int]] | None:
        """解析页码字符串，返回页码范围列表

        Args:
            pages_str: 形如 "1-,2,-3,4" 的页码字符串

        Returns:
            包含(start, end)元组的列表，其中-1表示无限制
        """
        if not pages_str:
            return None

        ranges = []
        for part in pages_str.split(","):
            part = part.strip()
            if "-" in part:
                start, end = part.split("-")
                start = int(start) if start else 1
                end = int(end) if end else -1
                ranges.append((start, end))
            else:
                page = int(part)
                ranges.append((page, page))
        return ranges

    def should_translate_page(self, page_number: int) -> bool:
        """判断指定页码是否需要翻译

        Args:
            page_number: 页码

        Returns:
            是否需要翻译该页
        """
        if not self.page_ranges:
            return True

        for start, end in self.page_ranges:
            if start <= page_number and (end == -1 or page_number <= end):
                return True
        return False



import pymupdf

from yadt.document_il import (
    Box,
    PdfCharacter,
    PdfLine,
    PdfParagraph,
    PdfParagraphComposition,
    PdfStyle,
    il_version_1,
    PdfFormula,
)
from yadt.document_il.utils.layout_helper import (
    get_paragraph_length_except,
    get_paragraph_max_height,
    get_paragraph_unicode,
)


class TypesettingUnit:
    def __init__(
        self,
        char: PdfCharacter = None,
        formular: PdfFormula = None,
        unicode: str = None,
        font: pymupdf.Font = None,
        font_size: float = None,
    ):
        assert (
            sum((x is None for x in [char, formular, unicode])) == 1
        ), "Only one of chars and formular can be not None"
        self.char = char
        self.formular = formular
        self.unicode = unicode

        if unicode:
            assert font_size, "Font size must be provided when unicode is provided"
            assert font, "Font must be provided when unicode is provided"
            assert len(unicode) == 1, "Unicode must be a single character"

            self.font = font
            self.font_size = font_size

    @property
    def box(self):
        if self.char:
            return self.char.box
        elif self.formular:
            return self.formular.box
        elif self.unicode:
            char_width = self.font.char_lengths(self.unicode, self.font_size)[0]
            return Box(0, 0, char_width, self.font_size)

    @property
    def width(self):
        return self.box.x2 - self.box.x

    @property
    def height(self):
        return self.box.y2 - self.box.y


class Typesetting:
    def __init__(self, font_path: str):
        self.font = pymupdf.Font(fontfile=font_path)

    def _typeset_pdf_character(
        self,
        char: il_version_1.PdfCharacter,
        current_x: float,
        current_y: float,
        scale: float,
    ) -> tuple[None, float] | tuple[PdfCharacter, float | None]:
        """排版一个已有的 PdfCharacter

        Args:
            char: 原始字符
            current_x: 当前 x 坐标

        Returns:
            新的 PdfCharacter 和新的 x 坐标
        """
        if char.char_unicode == " " and char.pdf_character_id is None:
            return None, current_x + (char.box.x2 - char.box.x) * scale
        new_char = il_version_1.PdfCharacter(
            pdf_character_id=char.pdf_character_id,
            char_unicode=char.char_unicode,
            box=il_version_1.Box(
                x=current_x,
                y=current_y,
                x2=current_x + (char.box.x2 - char.box.x) * scale,
                y2=current_y + (char.box.y2 - char.box.y) * scale,
            ),
            pdf_style=PdfStyle(
                font_id=char.pdf_style.font_id,
                font_size=char.pdf_style.font_size * scale,
                graphic_state=char.pdf_style.graphic_state,
            ),
            scale=scale,
            vertical=char.vertical,
            advance=char.advance * scale,
        )
        return new_char, new_char.box.x2

    def _typeset_unicode_char(
        self,
        char_unicode: str,
        x: float,
        y: float,
        font_size: float,
        font: pymupdf.Font,
        scale: float,
    ) -> tuple[il_version_1.PdfCharacter, float]:
        """排版一个 Unicode 字符

        Args:
            char_unicode: Unicode 字符
            x: x 坐标
            y: y 坐标
            font_size: 缩放前的字体大小
            font: 字体
            scale: 缩放比例
            style: 图形状态

        Returns:
            新的 PdfCharacter 和新的 x 坐标
        """
        font_size = font_size * scale
        char_width = font.char_lengths(char_unicode, font_size)[0]
        new_char = il_version_1.PdfCharacter(
            pdf_character_id=font.has_glyph(ord(char_unicode)),
            char_unicode=char_unicode,
            box=il_version_1.Box(
                x=x,
                y=y,
                x2=x + char_width,
                y2=y + font_size,
            ),
            pdf_style=PdfStyle(
                font_id="noto",
                font_size=font_size,
            ),
            scale=scale,
            vertical=False,
            advance=char_width,
        )
        return new_char, new_char.box.x2

    def _typeset_formula(
        self,
        formula: il_version_1.PdfFormula,
        noto_font: pymupdf.Font,
        x: float,
        y: float,
        scale: float,
    ) -> tuple[list[il_version_1.PdfCharacter], float]:
        """处理公式，保持公式内部字符的相对位置关系，整体缩放后放置到新位置

        Args:
            formula: 原始公式
            noto_font: 字体
            current_font_size: 当前字体大小
            x: 目标位置左下角x坐标
            y: 目标位置左下角y坐标
            scale: 缩放比例

        Returns:
            处理后的 PdfCharacter 列表
        """
        result_chars = []

        # 计算原始公式的边界框
        min_x = min(char.box.x for char in formula.pdf_character)
        min_y = min(char.box.y for char in formula.pdf_character)

        # 对每个字符进行缩放和位置调整
        for char in formula.pdf_character:
            if char.pdf_character_id is None:
                continue
            # 计算字符相对于公式左下角的偏移量
            relative_x = char.box.x - min_x
            relative_y = char.box.y - min_y  # PDF坐标系中y轴从下往上
            relative_width = char.box.x2 - char.box.x
            relative_height = char.box.y2 - char.box.y

            # 计算缩放后的新位置和大小
            new_x = x + relative_x * scale + formula.x_offset
            new_y = y + relative_y * scale + formula.y_offset
            new_width = relative_width * scale
            new_height = relative_height * scale

            # 创建新的字符对象
            new_char = il_version_1.PdfCharacter(
                char_unicode=char.char_unicode,
                box=il_version_1.Box(
                    x=new_x,
                    y=new_y,
                    x2=new_x + new_width,
                    y2=new_y + new_height,
                ),
                pdf_style=char.pdf_style,
                vertical=char.vertical,
                scale=scale,
                advance=char.advance,
                pdf_character_id=char.pdf_character_id,
            )
            result_chars.append(new_char)

        return result_chars, result_chars[-1].box.x2

    def process_toc_dots(
        self,
        paragraph: PdfParagraph,
        noto_font: pymupdf.Font,
        current_font_size: float,
        max_width: float,
        scale: float,
    ) -> list[il_version_1.PdfCharacter] | None:
        """处理目录条目的点号

        Args:
            paragraph: 原始段落
            noto_font: 字体
            current_font_size: 当前字体大小
            max_width: 最大可用宽度

        Returns:
            处理后的 PdfCharacter 列表，如果点号数量不足则返回 None
        """
        # 分割文本为标题和页码部分
        text = get_paragraph_unicode(paragraph)
        import re

        match = re.match(r"^(.*?)\s*\.+\s*(\d+)\s*$", text.strip())
        if not match:
            return None

        title, page_num = match.groups()
        title = title.strip()
        page_num = page_num.strip()

        # 计算除了点号以外的内容长度
        length_except_dots = get_paragraph_length_except(paragraph, ".", noto_font)

        # 计算点号的宽度
        dot_width = noto_font.char_lengths(".", current_font_size)[0]

        # 计算需要的点号数量
        dots_needed = max(1, int((max_width - length_except_dots) / dot_width))

        # 如果点号数量太少，返回 None
        if dots_needed < 5:
            return None

        result_chars = []
        current_x = 0

        # 复制标题部分的 PdfCharacter，遇到点号就停止
        for comp in paragraph.pdf_paragraph_composition:
            if comp.pdf_character:
                char = comp.pdf_character
                if char.char_unicode == ".":
                    break
                new_char, current_x = self._typeset_pdf_character(
                    char, current_x, paragraph.box.y, scale
                )
                if new_char:
                    result_chars.append(new_char)
            elif comp.pdf_same_style_characters:
                should_break = False
                for char in comp.pdf_same_style_characters.pdf_character:
                    if char.char_unicode == ".":
                        should_break = True
                        break
                    new_char, current_x = self._typeset_pdf_character(
                        char, current_x, paragraph.box.y, scale
                    )
                    if new_char:
                        result_chars.append(new_char)
                if should_break:
                    break
            elif comp.pdf_same_style_unicode_characters:
                should_break = False
                for char_unicode in comp.pdf_same_style_unicode_characters.unicode:
                    if char_unicode == ".":
                        should_break = True
                        break
                    new_char, current_x = self._typeset_unicode_char(
                        char_unicode,
                        current_x,
                        paragraph.box.y,
                        current_font_size,
                        noto_font,
                        scale,
                    )
                    result_chars.append(new_char)
                if should_break:
                    break
            elif comp.pdf_line:
                for char in comp.pdf_line.pdf_character:
                    if char.char_unicode == ".":
                        break
                    new_char, current_x = self._typeset_pdf_character(
                        char, current_x, paragraph.box.y, scale
                    )
                    if new_char:
                        result_chars.append(new_char)
            elif comp.pdf_formula:
                chars, current_x = self._typeset_formula(
                    comp.pdf_formula,
                    noto_font,
                    current_x,
                    paragraph.box.y,
                    scale,
                )
                result_chars.extend(chars)
            else:
                raise Exception(
                    "Unexpected composition type"
                    " in PdfParagraphComposition. "
                    "This type only appears in the IL "
                    "after the translation is completed."
                )

        # 添加点号
        for _ in range(dots_needed):
            new_char, current_x = self._typeset_unicode_char(
                ".", current_x, paragraph.box.y, current_font_size, noto_font, scale
            )
            result_chars.append(new_char)

        # TODO:想办法也透明传输页码？
        # 添加页码
        for char in page_num:
            new_char, current_x = self._typeset_unicode_char(
                char, current_x, paragraph.box.y, current_font_size, noto_font, scale
            )
            result_chars.append(new_char)

        return result_chars

    def try_typeset(
        self,
        paragraph: il_version_1.PdfParagraph,
        noto_font: pymupdf.Font,
        box: il_version_1.Box,
        scale: float,
        line_height: float,
    ) -> list[il_version_1.PdfCharacter] | None:
        """尝试对段落进行排版

        Args:
            paragraph: 要排版的段落
            noto_font: 字体
            box: 可用区域
            scale: 缩放比例

        Returns:
            排版后的字符列表，如果无法排版则返回 None
        """
        paragraph_max_height = get_paragraph_max_height(paragraph)
        result_chars = []
        current_x = box.x
        current_y = box.y2 - paragraph_max_height  # 从顶部开始排版
        line_height = paragraph_max_height * line_height  # 行高为字体大小的1.4倍

        # 遍历段落中的所有组成部分
        for comp in paragraph.pdf_paragraph_composition:
            if comp is None:
                continue
            if comp.pdf_character:
                char = comp.pdf_character
                char_width = char.box.x2 - char.box.x
                # 检查是否需要换行
                if current_x + char_width * scale > box.x2:
                    current_x = box.x
                    current_y -= line_height
                    # 检查是否超出底部边界
                    if current_y < box.y:
                        return None
                new_char, current_x = self._typeset_pdf_character(
                    char, current_x, current_y, scale
                )
                if new_char:
                    result_chars.append(new_char)
            elif comp.pdf_same_style_characters:
                for char in comp.pdf_same_style_characters.pdf_character:
                    char_width = char.box.x2 - char.box.x
                    if current_x + char_width * scale > box.x2:
                        current_x = box.x
                        current_y -= line_height
                        if current_y < box.y:
                            return None
                    new_char, current_x = self._typeset_pdf_character(
                        char, current_x, current_y, scale
                    )
                    if new_char:
                        result_chars.append(new_char)
            elif comp.pdf_same_style_unicode_characters:
                for char_unicode in comp.pdf_same_style_unicode_characters.unicode:
                    font_size = (
                        comp.pdf_same_style_unicode_characters.pdf_style.font_size
                        * scale
                    )
                    char_width = noto_font.char_lengths(char_unicode, font_size)[0]
                    if current_x + char_width * scale > box.x2:
                        current_x = box.x
                        current_y -= line_height
                        if current_y < box.y:
                            return None
                    new_char, current_x = self._typeset_unicode_char(
                        char_unicode,
                        current_x,
                        current_y,
                        font_size,
                        noto_font,
                        scale,
                    )
                    result_chars.append(new_char)
            elif comp.pdf_line:
                for char in comp.pdf_line.pdf_character:
                    char_width = char.box.x2 - char.box.x
                    if current_x + char_width * scale > box.x2:
                        current_x = box.x
                        current_y -= line_height
                        if current_y < box.y:
                            return None
                    new_char, current_x = self._typeset_pdf_character(
                        char, current_x, current_y, scale
                    )
                    if new_char:
                        result_chars.append(new_char)
            elif comp.pdf_formula:
                # 计算公式宽度
                formula_width = (
                    comp.pdf_formula.box.x2 - comp.pdf_formula.box.x
                ) * scale
                # 如果公式太宽，换到下一行
                if current_x + formula_width > box.x2:
                    current_x = box.x
                    current_y -= line_height
                    if current_y < box.y:
                        return None
                chars, current_x = self._typeset_formula(
                    comp.pdf_formula, noto_font, current_x, current_y, scale
                )
                result_chars.extend(chars)
            else:
                raise Exception(
                    "Unexpected composition type"
                    " in PdfParagraphComposition. "
                    "This type only appears in the IL "
                    "after the translation is completed."
                )

        return result_chars

    def get_max_right_space(self, current_box: Box, page) -> float:
        """获取段落右侧最大可用空间

        Args:
            current_box: 当前段落的边界框
            page: 当前页面

        Returns:
            可以扩展到的最大 x 坐标
        """
        # TODO: try to find right margin of page
        # 获取页面的裁剪框作为初始最大限制
        max_x = page.cropbox.box.x2 * 0.9

        # 检查所有可能的阻挡元素
        for para in page.pdf_paragraph:
            if para.box == current_box:  # 跳过当前段落
                continue
            # 只考虑在当前段落右侧且有垂直重叠的元素
            if para.box.x > current_box.x and not (
                para.box.y >= current_box.y2 or para.box.y2 <= current_box.y
            ):
                max_x = min(max_x, para.box.x)

        # 检查图形
        for figure in page.pdf_figure:
            if figure.box.x > current_box.x and not (
                figure.box.y >= current_box.y2 or figure.box.y2 <= current_box.y
            ):
                max_x = min(max_x, figure.box.x)

        return max_x

    def typsetting_document(self, document: il_version_1.Document):
        for page in document.page:
            # 开始实际的渲染过程
            for paragraph in page.pdf_paragraph:
                try:
                    self.render_paragraph_unicode_to_char(paragraph, self.font, 0.67)
                except ValueError:
                    # 获取段落当前的边界框
                    current_box = paragraph.box
                    # 获取右侧最大可用空间
                    max_x = self.get_max_right_space(current_box, page)
                    # 只有当有额外空间时才扩展
                    if max_x > current_box.x2:
                        expanded_box = Box(
                            x=current_box.x,
                            y=current_box.y,
                            x2=max_x,  # 直接扩展到最大可用位置
                            y2=current_box.y2,
                        )
                        # 更新段落的边界框
                        paragraph.box = expanded_box

                        # 重新渲染
                        self.render_paragraph_unicode_to_char(paragraph, self.font, 0.1)

    def render_paragraph_unicode_to_char(
        self,
        paragraph: il_version_1.PdfParagraph,
        noto_font: pymupdf.Font,
        scale_threshold,
    ):
        if not paragraph.pdf_paragraph_composition:
            return
        scale = 1.0
        # 尝试排版，如果失败则逐步缩小字号
        while scale >= scale_threshold:
            for line_hight in [1.7, 1.6, 1.5, 1.4, 1.3, 1.2, 1.1]:
                result = self.try_typeset(
                    paragraph, noto_font, paragraph.box, scale, line_hight
                )
                if result is not None:
                    paragraph.pdf_paragraph_composition = [
                        PdfParagraphComposition(pdf_character=char) for char in result
                    ]
                    paragraph.scale = scale
                    return
            if scale == 1.0:
                scale = 0.8
            else:
                scale -= 0.01
        raise ValueError(
            f"无法在保持最小缩放比 {scale_threshold} 的情况下排版文本。当前文本：{paragraph.unicode}"
        )



import base64
import math
import re
import unicodedata

from yadt.document_il.il_version_1 import (
    Box,
    Document,
    GraphicState,
    Page,
    PdfCharacter,
    PdfFormula,
    PdfLine,
    PdfParagraphComposition,
    PdfSameStyleCharacters,
    PdfStyle,
)
from yadt.document_il.utils.layout_helper import (
    get_char_unicode_string,
    is_same_style,
)
from yadt.translation_config import TranslationConfig


class StylesAndFormulas:
    def __init__(self, translation_config: TranslationConfig):
        self.translation_config = translation_config

    def process(self, document: Document):
        for page in document.page:
            self.process_page(page)

    def process_page(self, page: Page):
        """处理页面，包括公式识别和偏移量计算"""
        self.process_page_formulas(page)
        self.process_page_offsets(page)
        self.process_page_styles(page)

    def update_line_data(self, line: PdfLine):
        min_x = min(char.box.x for char in line.pdf_character)
        min_y = min(char.box.y for char in line.pdf_character)
        max_x = max(char.box.x2 for char in line.pdf_character)
        max_y = max(char.box.y2 for char in line.pdf_character)
        line.box = Box(min_x, min_y, max_x, max_y)

    def process_page_formulas(self, page: Page):
        if not page.pdf_paragraph:
            return

        # 收集该页所有的公式字体ID
        formula_font_ids = set()
        for font in page.pdf_font:
            if self.is_formulas_font(font.name):
                formula_font_ids.add(font.font_id)

        for paragraph in page.pdf_paragraph:
            if not paragraph.pdf_paragraph_composition:
                continue

            new_compositions = []
            current_chars = []
            is_current_formula = False  # 当前是否在处理公式字符

            for composition in paragraph.pdf_paragraph_composition:
                if not composition.pdf_line:
                    if current_chars:
                        # 处理剩余字符
                        new_compositions.append(
                            self.create_composition(
                                current_chars, is_current_formula
                            )
                        )
                        current_chars = []
                    new_compositions.append(composition)
                    continue

                line = composition.pdf_line
                for char in line.pdf_character:
                    is_formula = (
                        self.is_formulas_char(char.char_unicode)  # 公式字符
                        or char.pdf_style.font_id
                        in formula_font_ids  # 公式字体
                        or char.vertical  # 垂直字体
                        or (
                            len(current_chars) > 0
                            and not get_char_unicode_string(
                                current_chars
                            ).isspace()
                            # 角标字体，有 0.76 的角标和 0.799 的大写，这里用 0.79 取中，同时考虑首字母放大的情况
                            and char.pdf_style.font_size
                            < current_chars[-1].pdf_style.font_size * 0.79
                        )
                    )

                    if char.char_unicode == ' ':
                        is_formula = is_current_formula

                    if is_formula != is_current_formula and current_chars:
                        # 字符类型发生切换，处理之前的字符
                        new_compositions.append(
                            self.create_composition(
                                current_chars, is_current_formula
                            )
                        )
                        current_chars = []
                    is_current_formula = is_formula

                    current_chars.append(char)

                # 处理行末的字符
                if current_chars:
                    new_compositions.append(
                        self.create_composition(
                            current_chars, is_current_formula
                        )
                    )
                    current_chars = []

            paragraph.pdf_paragraph_composition = new_compositions

    def process_page_styles(self, page: Page):
        """处理页面中的文本样式，识别相同样式的文本"""
        if not page.pdf_paragraph:
            return

        for paragraph in page.pdf_paragraph:
            if not paragraph.pdf_paragraph_composition:
                continue

            # 计算基准样式（除公式外所有文字样式的交集）
            base_style = self._calculate_base_style(paragraph)
            paragraph.pdf_style = base_style

            # 重新组织段落中的文本，将相同样式的文本组合在一起
            new_compositions = []
            current_chars = []
            current_style = None

            for comp in paragraph.pdf_paragraph_composition:
                if comp.pdf_formula is not None:
                    if current_chars:
                        new_comp = self._create_same_style_composition(
                            current_chars, current_style
                        )
                        new_compositions.append(new_comp)
                        current_chars = []
                    new_compositions.append(comp)
                    continue

                if not comp.pdf_line:
                    continue

                for char in comp.pdf_line.pdf_character:
                    char_style = char.pdf_style
                    if current_style is None:
                        current_style = char_style
                        current_chars.append(char)
                    elif is_same_style(char_style, current_style):
                        current_chars.append(char)
                    else:
                        if current_chars:
                            new_comp = self._create_same_style_composition(
                                current_chars, current_style
                            )
                            new_compositions.append(new_comp)
                        current_chars = [char]
                        current_style = char_style

            if current_chars:
                new_comp = self._create_same_style_composition(
                    current_chars, current_style
                )
                new_compositions.append(new_comp)

            paragraph.pdf_paragraph_composition = new_compositions

    def _calculate_base_style(self, paragraph) -> PdfStyle:
        """计算段落的基准样式（除公式外所有文字样式的交集）"""
        styles = []
        for comp in paragraph.pdf_paragraph_composition:
            if isinstance(comp, PdfFormula):
                continue
            if not comp.pdf_line:
                continue
            for char in comp.pdf_line.pdf_character:
                styles.append(char.pdf_style)

        if not styles:
            return None

        # 返回所有样式的交集
        base_style = styles[0]
        for style in styles[1:]:
            # 更新基准样式为所有样式的交集
            base_style = self._merge_styles(base_style, style)

        # 如果font_id或font_size为None，则使用众数
        if base_style.font_id is None:
            base_style.font_id = self._get_mode_value(
                [s.font_id for s in styles])
        if base_style.font_size is None:
            base_style.font_size = self._get_mode_value(
                [s.font_size for s in styles]
            )

        return base_style

    def _get_mode_value(self, values):
        """计算列表中的众数"""
        if not values:
            return None
        from collections import Counter
        counter = Counter(values)
        return counter.most_common(1)[0][0]

    def _merge_styles(self, style1, style2):
        """合并两个样式，返回它们的交集"""
        if style1 is None or style1.font_size is None:
            return style2
        if style2 is None or style2.font_size is None:
            return style1

        return PdfStyle(
            font_id=style1.font_id
            if style1.font_id == style2.font_id
            else None,
            font_size=style1.font_size
            if math.fabs(style1.font_size - style2.font_size) < 0.02
            else None,
            graphic_state=self._merge_graphic_states(
                style1.graphic_state, style2.graphic_state
            ),
        )

    def _merge_graphic_states(self, state1, state2):
        """合并两个GraphicState，返回它们的交集"""
        if state1 is None:
            return state2
        if state2 is None:
            return state1

        return GraphicState(
            linewidth=state1.linewidth
            if state1.linewidth == state2.linewidth
            else None,
            dash=state1.dash if state1.dash == state2.dash else None,
            flatness=state1.flatness
            if state1.flatness == state2.flatness
            else None,
            intent=state1.intent if state1.intent == state2.intent else None,
            linecap=state1.linecap
            if state1.linecap == state2.linecap
            else None,
            linejoin=state1.linejoin
            if state1.linejoin == state2.linejoin
            else None,
            miterlimit=state1.miterlimit
            if state1.miterlimit == state2.miterlimit
            else None,
            ncolor=state1.ncolor if state1.ncolor == state2.ncolor else None,
            scolor=state1.scolor if state1.scolor == state2.scolor else None,
            stroking_color_space_name=state1.stroking_color_space_name
            if state1.stroking_color_space_name
            == state2.stroking_color_space_name
            else None,
            non_stroking_color_space_name=state1.non_stroking_color_space_name
            if state1.non_stroking_color_space_name
            == state2.non_stroking_color_space_name
            else None,
        )

    def _create_same_style_composition(
        self, chars: list[PdfCharacter], style
    ) -> PdfParagraphComposition:
        """创建具有相同样式的文本组合"""
        if not chars:
            return None

        # 计算边界框
        min_x = min(char.box.x for char in chars)
        min_y = min(char.box.y for char in chars)
        max_x = max(char.box.x2 for char in chars)
        max_y = max(char.box.y2 for char in chars)
        box = Box(min_x, min_y, max_x, max_y)

        return PdfParagraphComposition(
            pdf_same_style_characters=PdfSameStyleCharacters(
                box=box,
                pdf_style=style,
                pdf_character=chars,
            )
        )

    def process_page_offsets(self, page: Page):
        """计算公式的x和y偏移量"""
        if not page.pdf_paragraph:
            return

        for paragraph in page.pdf_paragraph:
            if not paragraph.pdf_paragraph_composition:
                continue

            # 计算该段落的行间距，用其80%作为容差
            line_spacing = self.calculate_line_spacing(paragraph)
            y_tolerance = line_spacing * 0.8

            for i, composition in enumerate(
                paragraph.pdf_paragraph_composition
            ):
                if not composition.pdf_formula:
                    continue

                formula = composition.pdf_formula
                left_line = None
                right_line = None

                # 查找左边最近的同一行的文本
                for j in range(i - 1, -1, -1):
                    comp = paragraph.pdf_paragraph_composition[j]
                    if comp.pdf_line:
                        # 检查y坐标是否接近，判断是否在同一行
                        if (
                            abs(comp.pdf_line.box.y - formula.box.y)
                            <= y_tolerance
                        ):
                            left_line = comp.pdf_line
                            break

                # 查找右边最近的同一行的文本
                for j in range(
                    i + 1, len(paragraph.pdf_paragraph_composition)
                ):
                    comp = paragraph.pdf_paragraph_composition[j]
                    if comp.pdf_line:
                        # 检查y坐标是否接近，判断是否在同一行
                        if (
                            abs(comp.pdf_line.box.y - formula.box.y)
                            <= y_tolerance
                        ):
                            right_line = comp.pdf_line
                            break

                # 计算x偏移量（相对于左边文本）
                if left_line:
                    formula.x_offset = formula.box.x - left_line.box.x2
                else:
                    formula.x_offset = 0  # 如果左边没有文字，x_offset应该为0
                if abs(formula.x_offset) < 0.1:
                    formula.x_offset = 0

                # 计算y偏移量
                if left_line:
                    # 使用底部坐标计算偏移量
                    formula.y_offset = formula.box.y - left_line.box.y
                elif right_line:
                    formula.y_offset = formula.box.y - right_line.box.y
                else:
                    formula.y_offset = 0

                if abs(formula.y_offset) < 0.1:
                    formula.y_offset = 0

    def calculate_line_spacing(self, paragraph) -> float:
        """计算段落中的平均行间距"""
        if not paragraph.pdf_paragraph_composition:
            return 0.0

        # 收集所有文本行的y坐标
        line_y_positions = []
        for comp in paragraph.pdf_paragraph_composition:
            if comp.pdf_line:
                line_y_positions.append(comp.pdf_line.box.y)

        if len(line_y_positions) < 2:
            return 10.0  # 如果只有一行或没有行，返回一个默认值

        # 计算相邻行之间的y差值
        line_spacings = []
        for i in range(len(line_y_positions) - 1):
            spacing = abs(line_y_positions[i] - line_y_positions[i + 1])
            if spacing > 0:  # 忽略重叠的行
                line_spacings.append(spacing)

        if not line_spacings:
            return 10.0  # 如果没有有效的行间距，返回默认值

        # 使用中位数来避免异常值的影响
        median_spacing = sorted(line_spacings)[len(line_spacings) // 2]
        return median_spacing

    def create_composition(
        self, chars: list[PdfCharacter], is_formula: bool
    ) -> PdfParagraphComposition:
        if is_formula:
            formula = PdfFormula(pdf_character=chars)
            self.update_formula_data(formula)
            return PdfParagraphComposition(pdf_formula=formula)
        else:
            new_line = PdfLine(pdf_character=chars)
            self.update_line_data(new_line)
            return PdfParagraphComposition(pdf_line=new_line)

    def update_formula_data(self, formula: PdfFormula):
        min_x = min(char.box.x for char in formula.pdf_character)
        min_y = min(char.box.y for char in formula.pdf_character)
        max_x = max(char.box.x2 for char in formula.pdf_character)
        max_y = max(char.box.y2 for char in formula.pdf_character)
        formula.box = Box(min_x, min_y, max_x, max_y)

    def is_formulas_font(self, font_name: str) -> bool:
        if self.translation_config.formular_font_pattern:
            pattern = self.translation_config.formular_font_pattern
        else:
            pattern = (
                r"(CM[^RB]"
                r"|(MS|XY|MT|BL|RM|EU|LA|RS)[A-Z]"
                r"|LINE"
                r"|LCIRCLE"
                r"|TeX-"
                r"|rsfs"
                r"|txsy"
                r"|wasy"
                r"|stmary"
                r"|.*Mono"
                r"|.*Code"
                r"|.*Ital"
                r"|.*Sym"
                r"|.*Math"
                r")"
            )

        if font_name.startswith("BASE64:"):
            font_name_bytes = base64.b64decode(font_name[7:])
            font = font_name_bytes.split(b"+")[-1]
            pattern = pattern.encode()
        else:
            font = font_name.split("+")[-1]

        if re.match(pattern, font):
            return True

        return False

    def is_formulas_char(self, char: str) -> bool:
        if self.translation_config.formular_char_pattern:
            pattern = self.translation_config.formular_char_pattern
            if re.match(pattern, char):
                return True
        if (
            char
            and char != " "  # 非空格
            and (
                unicodedata.category(char[0])
                in [
                    "Lm",
                    "Mn",
                    "Sk",
                    "Sm",
                    "Zl",
                    "Zp",
                    "Zs",
                ]  # 文字修饰符、数学符号、分隔符号
                or ord(char[0]) in range(0x370, 0x400)  # 希腊字母
            )
        ):
            return True

        return False



import base64

import numpy as np
import pdfminer.pdfinterp
import pymupdf
from pdfminer.layout import LTChar, LTFigure
from pdfminer.pdffont import PDFCIDFont, PDFFont

from yadt.doclayout import DocLayoutModel
from yadt.document_il import il_version_1
from yadt.translation_config import TranslationConfig


class ILCreater:
    def __init__(self, translation_config: TranslationConfig):
        self.current_page: il_version_1.Page = None
        self.mupdf: pymupdf.Document = None
        self.model = DocLayoutModel.load_available()
        self.docs = il_version_1.Document(page=[])
        self.stroking_color_space_name = None
        self.non_stroking_color_space_name = None
        self.translation_config = translation_config

    # pdf32000 page 171
    def on_stroking_color_space(self, color_space_name):
        self.stroking_color_space_name = color_space_name

    def on_non_stroking_color_space(self, color_space_name):
        self.non_stroking_color_space_name = color_space_name

    def on_page_start(self):
        self.current_page = il_version_1.Page(
            pdf_font=[],
            pdf_character=[],
            page_layout=[],
            # currently don't support UserUnit page parameter
            # pdf32000 page 79
            unit="point",
        )
        self.current_page_font_name_id_map = {}
        self.docs.page.append(self.current_page)

    def on_page_crop_box(
        self,
        x0: float | int,
        y0: float | int,
        x1: float | int,
        y1: float | int,
    ):
        box = il_version_1.Box(
            x=float(x0), y=float(y0), x2=float(x1), y2=float(y1)
        )
        self.current_page.cropbox = il_version_1.Cropbox(box=box)

    def on_page_media_box(
        self,
        x0: float | int,
        y0: float | int,
        x1: float | int,
        y1: float | int,
    ):
        box = il_version_1.Box(
            x=float(x0), y=float(y0), x2=float(x1), y2=float(y1)
        )
        self.current_page.mediabox = il_version_1.Mediabox(box=box)

    def on_page_number(self, page_number: int):
        assert isinstance(page_number, int)
        assert page_number >= 0
        self.current_page.page_number = page_number

        self.on_page_layout(page_number)

    def on_page_base_operation(self, operation: str):
        self.current_page.base_operations = il_version_1.BaseOperations(
            value=operation
        )

    def on_page_resource_font(self, font: PDFFont, xref_id: int, font_id: str):
        font_name = font.fontname
        if isinstance(font.fontname, bytes):
            try:
                font_name = font.fontname.decode("utf-8")
            except UnicodeDecodeError:
                font_name = "BASE64:" + base64.b64encode(font_name).decode(
                    "utf-8"
                )
        encoding_length = 1
        if isinstance(font, PDFCIDFont):
            encoding_length = 2
        il_font_metadata = il_version_1.PdfFont(
            name=font_name,
            xref_id=xref_id,
            font_id=font_id,
            encoding_length=encoding_length,
        )
        self.current_page_font_name_id_map[font_name] = font_id
        self.current_page.pdf_font.append(il_font_metadata)

    def create_graphic_state(self, gs: pdfminer.pdfinterp.PDFGraphicState):
        graphic_state = il_version_1.GraphicState()
        for k, v in gs.__dict__.items():
            if v is None:
                continue
            if k in ["scolor", "ncolor"]:
                if isinstance(v, tuple):
                    v = list(v)
                else:
                    v = [v]
                setattr(graphic_state, k, v)
                continue
            if k == "linewidth":
                graphic_state.linewidth = float(v)
                continue
            continue
            raise NotImplementedError

        graphic_state.stroking_color_space_name = (
            self.stroking_color_space_name
        )
        graphic_state.non_stroking_color_space_name = (
            self.non_stroking_color_space_name
        )
        return graphic_state

    def on_lt_char(self, char: LTChar):
        gs = self.create_graphic_state(char.graphicstate)
        bbox = il_version_1.Box(
            char.bbox[0], char.bbox[1], char.bbox[2], char.bbox[3]
        )

        font_id = self.current_page_font_name_id_map[char.font.fontname]
        char_id = char.cid
        char_unicode = char.get_text()
        advance = char.adv
        if char.matrix[0] == 0 and char.matrix[3] == 0:
            vertical = True
        else:
            vertical = False
        pdf_style = il_version_1.PdfStyle(
            font_id=font_id,
            font_size=char.size,
            graphic_state=gs,
        )
        pdf_char = il_version_1.PdfCharacter(
            box=bbox,
            pdf_character_id=char_id,
            advance=advance,
            char_unicode=char_unicode,
            vertical=vertical,
            pdf_style=pdf_style,
        )
        self.current_page.pdf_character.append(pdf_char)

    def on_page_layout(self, page_number):
        if (
            self.translation_config.should_translate_page(page_number + 1)
            is False
        ):
            return
        pix = self.mupdf[page_number].get_pixmap()
        image = np.fromstring(pix.samples, np.uint8).reshape(
            pix.height, pix.width, 3
        )[:, :, ::-1]
        h, w = pix.height, pix.width
        layouts = self.model.predict(image, imgsz=int(pix.height / 32) * 32)[0]
        id = 0
        for layout in layouts.boxes:
            id += 1
            # Convert the coordinate system from the picture coordinate system to the il coordinate system
            x0, y0, x1, y1 = layout.xyxy
            x0, y0, x1, y1 = (
                np.clip(int(x0 - 1), 0, w - 1),
                np.clip(int(h - y1 - 1), 0, h - 1),
                np.clip(int(x1 + 1), 0, w - 1),
                np.clip(int(h - y0 + 1), 0, h - 1),
            )
            page_layout = il_version_1.PageLayout(
                id=id,
                box=il_version_1.Box(
                    x0.item(), y0.item(), x1.item(), y1.item()
                ),
                conf=layout.conf.item(),
                class_name=layouts.names[layout.cls],
            )
            self.docs.page[page_number].page_layout.append(page_layout)

    def create_il(self):
        pages = [
            page
            for page in self.docs.page
            if self.translation_config.should_translate_page(
                page.page_number + 1
            )
        ]
        self.docs.page = pages
        return self.docs

    def on_total_pages(self, total_pages: int):
        assert isinstance(total_pages, int)
        assert total_pages > 0
        self.docs.total_pages = total_pages

    def on_pdf_figure(self, figure: LTFigure):
        box = il_version_1.Box(
            figure.bbox[0], figure.bbox[1], figure.bbox[2], figure.bbox[3]
        )
        self.current_page.pdf_figure.append(il_version_1.PdfFigure(box=box))



import html
import os
import re
import threading
import time
import unicodedata
from abc import ABC
from copy import copy

import openai
import requests

from yadt.document_il.translator.cache import TranslationCache


def remove_control_characters(s):
    return "".join(ch for ch in s if unicodedata.category(ch)[0] != "C")


class RateLimiter:
    def __init__(self, max_qps: int):
        self.max_qps = max_qps
        self.min_interval = 1.0 / max_qps
        self.last_requests = []  # Track last N requests
        self.window_size = max_qps  # Track requests in a sliding window
        self.lock = threading.Lock()

    def wait(self):
        with self.lock:
            now = time.time()

            # Clean up old requests outside the 1-second window
            while self.last_requests and now - self.last_requests[0] > 1.0:
                self.last_requests.pop(0)

            # If we have less than max_qps requests in the last second, allow immediately
            if len(self.last_requests) < self.max_qps:
                self.last_requests.append(now)
                return

            # Otherwise, wait until we can make the next request
            next_time = self.last_requests[0] + 1.0
            if next_time > now:
                time.sleep(next_time - now)
            self.last_requests.pop(0)
            self.last_requests.append(next_time)

    def set_max_qps(self, max_qps):
        self.max_qps = max_qps
        self.min_interval = 1.0 / max_qps
        self.window_size = max_qps


_translate_rate_limiter = RateLimiter(5)


def set_translate_rate_limiter(max_qps):
    _translate_rate_limiter.set_max_qps(max_qps)


class BaseTranslator(ABC):
    # Due to cache limitations, name should be within 20 characters.
    # cache.py: translate_engine = CharField(max_length=20)
    name = "base"
    lang_map = {}

    def __init__(self, lang_in, lang_out, ignore_cache):
        self.ignore_cache = ignore_cache
        lang_in = self.lang_map.get(lang_in.lower(), lang_in)
        lang_out = self.lang_map.get(lang_out.lower(), lang_out)
        self.lang_in = lang_in
        self.lang_out = lang_out

        self.cache = TranslationCache(
            self.name,
            {
                "lang_in": lang_in,
                "lang_out": lang_out,
            },
        )

        self.translate_call_count = 0
        self.translate_cache_call_count = 0

    def __del__(self):
        print(f"{self.name} translate call count: {self.translate_call_count}")
        print(
            f"{self.name} translate cache call count: {self.translate_cache_call_count}"
        )

    def add_cache_impact_parameters(self, k: str, v):
        """
        Add parameters that affect the translation quality to distinguish the translation effects under different parameters.
        :param k: key
        :param v: value
        """
        self.cache.add_params(k, v)

    def translate(self, text, ignore_cache=False):
        """
        Translate the text, and the other part should call this method.
        :param text: text to translate
        :return: translated text
        """
        self.translate_call_count += 1
        if not (self.ignore_cache or ignore_cache):
            cache = self.cache.get(text)
            if cache is not None:
                self.translate_cache_call_count += 1
                return cache
        _translate_rate_limiter.wait()
        translation = self.do_translate(text)
        if not (self.ignore_cache or ignore_cache):
            self.cache.set(text, translation)
        return translation

    def do_translate(self, text):
        """
        Actual translate text, override this method
        :param text: text to translate
        :return: translated text
        """
        raise NotImplementedError

    def __str__(self):
        return f"{self.name} {self.lang_in} {self.lang_out} {self.model}"

    def get_rich_text_left_placeholder(self, id: int):
        return f"<b{id}>"

    def get_rich_text_right_placeholder(self, id: int):
        return f"</b{id}>"

    def get_formular_placeholder(self, id: int):
        return self.get_rich_text_left_placeholder(
            id
        ) + self.get_rich_text_right_placeholder(id)


class GoogleTranslator(BaseTranslator):
    name = "google"
    lang_map = {"zh": "zh-CN"}

    def __init__(self, lang_in, lang_out, ignore_cache=False):
        super().__init__(lang_in, lang_out, ignore_cache)
        self.session = requests.Session()
        self.endpoint = "http://translate.google.com/m"
        self.headers = {
            "User-Agent": "Mozilla/4.0 (compatible;MSIE 6.0;Windows NT 5.1;SV1;.NET CLR 1.1.4322;.NET CLR 2.0.50727;.NET CLR 3.0.04506.30)"
            # noqa: E501
        }

    def do_translate(self, text):
        text = text[:5000]  # google translate max length
        response = self.session.get(
            self.endpoint,
            params={"tl": self.lang_out, "sl": self.lang_in, "q": text},
            headers=self.headers,
        )
        re_result = re.findall(
            r'(?s)class="(?:t0|result-container)">(.*?)<', response.text
        )
        if response.status_code == 400:
            result = "IRREPARABLE TRANSLATION ERROR"
        else:
            response.raise_for_status()
            result = html.unescape(re_result[0])
        return remove_control_characters(result)


class OpenAITranslator(BaseTranslator):
    # https://github.com/openai/openai-python
    name = "openai"

    def __init__(
        self,
        lang_in,
        lang_out,
        model,
        base_url=None,
        api_key=None,
        ignore_cache=False,
    ):
        super().__init__(lang_in, lang_out, ignore_cache)
        self.options = {"temperature": 0}  # 随机采样可能会打断公式标记
        self.client = openai.OpenAI(base_url=base_url, api_key=api_key)
        self.add_cache_impact_parameters(
            "temperature", self.options["temperature"]
        )
        self.model = model
        self.add_cache_impact_parameters("model", self.model)
        self.add_cache_impact_parameters("prompt", self.prompt(""))

    def do_translate(self, text) -> str:
        response = self.client.chat.completions.create(
            model=self.model,
            **self.options,
            messages=self.prompt(text),
        )
        return response.choices[0].message.content.strip()

    def prompt(self, text):
        return [
            {
                "role": "system",
                "content": "You are a professional,authentic machine translation engine.",
            },
            {
                "role": "user",
                "content": f";; Treat next line as plain text input and translate it into {self.lang_out}, output translation ONLY. If translation is unnecessary (e.g. proper nouns, codes, {'{{1}}, etc. '}), return the original text. NO explanations. NO notes. Input:\n\n{text}",
                # noqa: E501
            },
        ]

    def get_formular_placeholder(self, id: int):
        return "{{v" + str(id) + "}}"
        return "{{" + str(id) + "}}"

    def get_rich_text_left_placeholder(self, id: int):
        return self.get_formular_placeholder(id)

    def get_rich_text_right_placeholder(self, id: int):
        return self.get_formular_placeholder(id + 1)



import abc
import os.path

import cv2
import numpy as np
import ast
import onnx
import onnxruntime
from huggingface_hub import hf_hub_download


class DocLayoutModel(abc.ABC):
    @staticmethod
    def load_onnx():
        model = OnnxModel.from_pretrained(
            repo_id="wybxc/DocLayout-YOLO-DocStructBench-onnx",
            filename="doclayout_yolo_docstructbench_imgsz1024.onnx",
        )
        return model

    @staticmethod
    def load_available():
        return DocLayoutModel.load_onnx()

    @property
    @abc.abstractmethod
    def stride(self) -> int:
        """Stride of the model input."""
        pass

    @abc.abstractmethod
    def predict(self, image, imgsz=1024, **kwargs) -> list:
        """
        Predict the layout of a document page.

        Args:
            image: The image of the document page.
            imgsz: Resize the image to this size. Must be a multiple of the stride.
            **kwargs: Additional arguments.
        """
        pass


class YoloResult:
    """Helper class to store detection results from ONNX model."""

    def __init__(self, boxes, names):
        self.boxes = [YoloBox(data=d) for d in boxes]
        self.boxes.sort(key=lambda x: x.conf, reverse=True)
        self.names = names


class YoloBox:
    """Helper class to store detection results from ONNX model."""

    def __init__(self, data):
        self.xyxy = data[:4]
        self.conf = data[-2]
        self.cls = data[-1]


providers = [
    ('CoreMLExecutionProvider', {
        "ModelFormat": "MLProgram", "MLComputeUnits": "ALL",
        "RequireStaticInputShapes": "0", "EnableOnSubgraphs": "0"
    }),
]
class OnnxModel(DocLayoutModel):
    def __init__(self, model_path: str):
        self.model_path = model_path

        model = onnx.load(model_path)
        metadata = {d.key: d.value for d in model.metadata_props}
        self._stride = ast.literal_eval(metadata["stride"])
        self._names = ast.literal_eval(metadata["names"])

        self.model = onnxruntime.InferenceSession(model.SerializeToString(), providers=providers)

    @staticmethod
    def from_pretrained(repo_id: str, filename: str):
        if os.environ.get("USE_MODELSCOPE", "0") == "1":
            repo_mapping = {
                # Edit here to add more models
                "wybxc/DocLayout-YOLO-DocStructBench-onnx": "AI-ModelScope/DocLayout-YOLO-DocStructBench-onnx"
            }
            from modelscope import snapshot_download

            model_dir = snapshot_download(repo_mapping[repo_id])
            pth = os.path.join(model_dir, filename)
        else:
            pth = hf_hub_download(repo_id=repo_id, filename=filename, etag_timeout=1, local_files_only=True)
        return OnnxModel(pth)

    @property
    def stride(self):
        return self._stride

    def resize_and_pad_image(self, image, new_shape):
        """
        Resize and pad the image to the specified size, ensuring dimensions are multiples of stride.

        Parameters:
        - image: Input image
        - new_shape: Target size (integer or (height, width) tuple)
        - stride: Padding alignment stride, default 32

        Returns:
        - Processed image
        """
        if isinstance(new_shape, int):
            new_shape = (new_shape, new_shape)

        h, w = image.shape[:2]
        new_h, new_w = new_shape

        # Calculate scaling ratio
        r = min(new_h / h, new_w / w)
        resized_h, resized_w = int(round(h * r)), int(round(w * r))

        # Resize image
        image = cv2.resize(
            image, (resized_w, resized_h), interpolation=cv2.INTER_LINEAR
        )

        # Calculate padding size and align to stride multiple
        pad_w = (new_w - resized_w) % self.stride
        pad_h = (new_h - resized_h) % self.stride
        top, bottom = pad_h // 2, pad_h - pad_h // 2
        left, right = pad_w // 2, pad_w - pad_w // 2

        # Add padding
        image = cv2.copyMakeBorder(
            image, top, bottom, left, right, cv2.BORDER_CONSTANT, value=(114, 114, 114)
        )

        return image

    def scale_boxes(self, img1_shape, boxes, img0_shape):
        """
        Rescales bounding boxes (in the format of xyxy by default) from the shape of the image they were originally
        specified in (img1_shape) to the shape of a different image (img0_shape).

        Args:
            img1_shape (tuple): The shape of the image that the bounding boxes are for,
                in the format of (height, width).
            boxes (torch.Tensor): the bounding boxes of the objects in the image, in the format of (x1, y1, x2, y2)
            img0_shape (tuple): the shape of the target image, in the format of (height, width).

        Returns:
            boxes (torch.Tensor): The scaled bounding boxes, in the format of (x1, y1, x2, y2)
        """

        # Calculate scaling ratio
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])

        # Calculate padding size
        pad_x = round((img1_shape[1] - img0_shape[1] * gain) / 2 - 0.1)
        pad_y = round((img1_shape[0] - img0_shape[0] * gain) / 2 - 0.1)

        # Remove padding and scale boxes
        boxes[..., :4] = (boxes[..., :4] - [pad_x, pad_y, pad_x, pad_y]) / gain
        return boxes

    def predict(self, image, imgsz=1024, **kwargs):
        # Preprocess input image
        orig_h, orig_w = image.shape[:2]
        pix = self.resize_and_pad_image(image, new_shape=imgsz)
        pix = np.transpose(pix, (2, 0, 1))  # CHW
        pix = np.expand_dims(pix, axis=0)  # BCHW
        pix = pix.astype(np.float32) / 255.0  # Normalize to [0, 1]
        new_h, new_w = pix.shape[2:]

        # Run inference
        preds = self.model.run(None, {"images": pix})[0]

        # Postprocess predictions
        preds = preds[preds[..., 4] > 0.25]
        preds[..., :4] = self.scale_boxes(
            (new_h, new_w), preds[..., :4], (orig_h, orig_w)
        )
        return [YoloResult(boxes=preds, names=self._names)]



import logging
from typing import Any, Dict, Optional, Sequence, Tuple, cast
import numpy as np

from pdfminer import settings
from pdfminer.pdfcolor import PREDEFINED_COLORSPACE, PDFColorSpace
from pdfminer.pdfdevice import PDFDevice
from pdfminer.pdfinterp import (
    PDFPageInterpreter,
    PDFResourceManager,
    PDFContentParser,
    PDFInterpreterError,
    Color,
    PDFStackT,
    LITERAL_FORM,
    LITERAL_IMAGE,
)
from pdfminer.pdffont import PDFFont
from pdfminer.pdfpage import PDFPage
from pdfminer.pdftypes import (
    PDFObjRef,
    dict_value,
    list_value,
    resolve1,
    stream_value,
)
from pdfminer.psexceptions import PSEOF
from pdfminer.psparser import (
    PSKeyword,
    keyword_name,
    literal_name,
)
from pdfminer.utils import (
    MATRIX_IDENTITY,
    Matrix,
    Rect,
    mult_matrix,
    apply_matrix_pt,
)

from yadt.document_il.frontend.il_creater import ILCreater

log = logging.getLogger(__name__)


def safe_float(o: Any) -> Optional[float]:
    try:
        return float(o)
    except (TypeError, ValueError):
        return None


class PDFPageInterpreterEx(PDFPageInterpreter):
    """Processor for the content of a PDF page

    Reference: PDF Reference, Appendix A, Operator Summary
    """

    def __init__(
        self,
        rsrcmgr: PDFResourceManager,
        device: PDFDevice,
        obj_patch,
        il_creater: ILCreater,
    ) -> None:
        self.rsrcmgr = rsrcmgr
        self.device = device
        self.obj_patch = obj_patch
        self.il_creater = il_creater

    def dup(self) -> "PDFPageInterpreterEx":
        return self.__class__(self.rsrcmgr, self.device, self.obj_patch, self.il_creater)

    def init_resources(self, resources: Dict[object, object]) -> None:
        # 重载设置 fontid 和 descent
        """Prepare the fonts and XObjects listed in the Resource attribute."""
        self.resources = resources
        self.fontmap: Dict[object, PDFFont] = {}
        self.fontid: Dict[PDFFont, object] = {}
        self.xobjmap = {}
        self.csmap: Dict[str, PDFColorSpace] = PREDEFINED_COLORSPACE.copy()
        if not resources:
            return

        def get_colorspace(spec: object) -> Optional[PDFColorSpace]:
            if isinstance(spec, list):
                name = literal_name(spec[0])
            else:
                name = literal_name(spec)
            if name == "ICCBased" and isinstance(spec, list) and len(spec) >= 2:
                return PDFColorSpace(name, stream_value(spec[1])["N"])
            elif name == "DeviceN" and isinstance(spec, list) and len(spec) >= 2:
                return PDFColorSpace(name, len(list_value(spec[1])))
            else:
                return PREDEFINED_COLORSPACE.get(name)

        for k, v in dict_value(resources).items():
            # log.debug("Resource: %r: %r", k, v)
            if k == "Font":
                for fontid, spec in dict_value(v).items():
                    objid = None
                    if isinstance(spec, PDFObjRef):
                        objid = spec.objid
                    spec = dict_value(spec)
                    font = self.rsrcmgr.get_font(objid, spec)
                    self.il_creater.on_page_resource_font(font, objid, fontid)
                    self.fontmap[fontid] = font
                    self.fontmap[fontid].descent = 0  # hack fix descent
                    self.fontid[self.fontmap[fontid]] = fontid
            elif k == "ColorSpace":
                for csid, spec in dict_value(v).items():
                    colorspace = get_colorspace(resolve1(spec))
                    if colorspace is not None:
                        self.csmap[csid] = colorspace
            elif k == "ProcSet":
                self.rsrcmgr.get_procset(list_value(v))
            elif k == "XObject":
                for xobjid, xobjstrm in dict_value(v).items():
                    self.xobjmap[xobjid] = xobjstrm
        pass

    def do_S(self) -> None:
        # 重载过滤非公式线条
        """Stroke path"""

        def is_black(color: Color) -> bool:
            if isinstance(color, Tuple):
                return sum(color) == 0
            else:
                return color == 0

        if (
            len(self.curpath) == 2
            and self.curpath[0][0] == "m"
            and self.curpath[1][0] == "l"
            and apply_matrix_pt(self.ctm, self.curpath[0][-2:])[1]
            == apply_matrix_pt(self.ctm, self.curpath[1][-2:])[1]
            and is_black(self.graphicstate.scolor)
        ):  # 独立直线，水平，黑色
            # print(apply_matrix_pt(self.ctm,self.curpath[0][-2:]),apply_matrix_pt(self.ctm,self.curpath[1][-2:]),self.graphicstate.scolor)
            self.device.paint_path(self.graphicstate, True, False, False, self.curpath)
            self.curpath = []
            return "n"
        else:
            self.curpath = []

    def do_CS(self, name: PDFStackT) -> None:
        """Set color space for stroking operations

        Introduced in PDF 1.1
        """
        try:
            self.il_creater.on_stroking_color_space(literal_name(name))
            self.scs = self.csmap[literal_name(name)]
        except KeyError:
            if settings.STRICT:
                raise PDFInterpreterError("Undefined ColorSpace: %r" % name)
        return

    def do_cs(self, name: PDFStackT) -> None:
        """Set color space for nonstroking operations"""
        try:
            self.il_creater.on_non_stroking_color_space(literal_name(name))
            self.ncs = self.csmap[literal_name(name)]
        except KeyError:
            if settings.STRICT:
                raise PDFInterpreterError("Undefined ColorSpace: %r" % name)
        return

    ############################################################
    # 重载过滤非公式线条（F/B）
    def do_f(self) -> None:
        """Fill path using nonzero winding number rule"""
        # self.device.paint_path(self.graphicstate, False, True, False, self.curpath)
        self.curpath = []

    def do_F(self) -> None:
        """Fill path using nonzero winding number rule (obsolete)"""

    def do_f_a(self) -> None:
        """Fill path using even-odd rule"""
        # self.device.paint_path(self.graphicstate, False, True, True, self.curpath)
        self.curpath = []

    def do_B(self) -> None:
        """Fill and stroke path using nonzero winding number rule"""
        # self.device.paint_path(self.graphicstate, True, True, False, self.curpath)
        self.curpath = []

    def do_B_a(self) -> None:
        """Fill and stroke path using even-odd rule"""
        # self.device.paint_path(self.graphicstate, True, True, True, self.curpath)
        self.curpath = []

    ############################################################
    # 重载返回调用参数（SCN）
    def do_SCN(self) -> None:
        """Set color for stroking operations."""
        if self.scs:
            n = self.scs.ncomponents
        else:
            if settings.STRICT:
                raise PDFInterpreterError("No colorspace specified!")
            n = 1
        args = self.pop(n)
        self.graphicstate.scolor = cast(Color, args)
        return args

    def do_scn(self) -> None:
        """Set color for nonstroking operations"""
        if self.ncs:
            n = self.ncs.ncomponents
        else:
            if settings.STRICT:
                raise PDFInterpreterError("No colorspace specified!")
            n = 1
        args = self.pop(n)
        self.graphicstate.ncolor = cast(Color, args)
        return args

    def do_SC(self) -> None:
        """Set color for stroking operations"""
        return self.do_SCN()

    def do_sc(self) -> None:
        """Set color for nonstroking operations"""
        return self.do_scn()

    def do_Do(self, xobjid_arg: PDFStackT) -> None:
        # 重载设置 xobj 的 obj_patch
        """Invoke named XObject"""
        xobjid = literal_name(xobjid_arg)
        try:
            xobj = stream_value(self.xobjmap[xobjid])
        except KeyError:
            if settings.STRICT:
                raise PDFInterpreterError("Undefined xobject id: %r" % xobjid)
            return
        # log.debug("Processing xobj: %r", xobj)
        subtype = xobj.get("Subtype")
        if subtype is LITERAL_FORM and "BBox" in xobj:
            interpreter = self.dup()
            bbox = cast(Rect, list_value(xobj["BBox"]))
            matrix = cast(Matrix, list_value(xobj.get("Matrix", MATRIX_IDENTITY)))
            # According to PDF reference 1.7 section 4.9.1, XObjects in
            # earlier PDFs (prior to v1.2) use the page's Resources entry
            # instead of having their own Resources entry.
            xobjres = xobj.get("Resources")
            if xobjres:
                resources = dict_value(xobjres)
            else:
                resources = self.resources.copy()
            self.device.begin_figure(xobjid, bbox, matrix)
            ctm = mult_matrix(matrix, self.ctm)
            ops_base = interpreter.render_contents(
                resources,
                [xobj],
                ctm=ctm,
            )
            try:  # 有的时候 form 字体加不上这里会烂掉
                self.device.fontid = interpreter.fontid
                self.device.fontmap = interpreter.fontmap
                ops_new = self.device.end_figure(xobjid)
                ctm_inv = np.linalg.inv(np.array(ctm[:4]).reshape(2, 2))
                np_version = np.__version__
                if np_version.split(".")[0] >= "2":
                    pos_inv = -np.asmatrix(ctm[4:]) * ctm_inv
                else:
                    pos_inv = -np.mat(ctm[4:]) * ctm_inv
                a, b, c, d = ctm_inv.reshape(4).tolist()
                e, f = pos_inv.tolist()[0]
                self.obj_patch[self.xobjmap[xobjid].objid] = (
                    f"q {ops_base}Q {a} {b} {c} {d} {e} {f} cm {ops_new}"
                )
            except Exception:
                pass
        elif subtype is LITERAL_IMAGE and "Width" in xobj and "Height" in xobj:
            self.device.begin_figure(xobjid, (0, 0, 1, 1), MATRIX_IDENTITY)
            self.device.render_image(xobjid, xobj)
            self.device.end_figure(xobjid)
        else:
            # unsupported xobject type.
            pass

    def process_page(self, page: PDFPage) -> None:
        # 重载设置 page 的 obj_patch
        # log.debug("Processing page: %r", page)
        # print(page.mediabox,page.cropbox)
        # (x0, y0, x1, y1) = page.mediabox
        (x0, y0, x1, y1) = page.cropbox
        if page.rotate == 90:
            ctm = (0, -1, 1, 0, -y0, x1)
        elif page.rotate == 180:
            ctm = (-1, 0, 0, -1, x1, y1)
        elif page.rotate == 270:
            ctm = (0, 1, -1, 0, y1, -x0)
        else:
            ctm = (1, 0, 0, 1, -x0, -y0)
        self.il_creater.on_page_start()
        self.il_creater.on_page_crop_box(x0, y0, x1, y1)
        self.device.begin_page(page, ctm)
        ops_base = self.render_contents(page.resources, page.contents, ctm=ctm)
        self.device.fontid = self.fontid
        self.device.fontmap = self.fontmap
        ops_new = self.device.end_page(page)
        # 上面渲染的时候会根据 cropbox 减掉页面偏移得到真实坐标，这里输出的时候需要用 cm 把页面偏移加回来
        self.obj_patch[page.page_xref] = (
            # f"q {ops_base}Q 1 0 0 1 {x0} {y0} cm {ops_new}"  # ops_base 里可能有图，需要让 ops_new 里的文字覆盖在上面，使用 q/Q 重置位置矩阵
            ""
        )
        for obj in page.contents:
            self.obj_patch[obj.objid] = ""
        return ops_base

    def render_contents(
        self,
        resources: Dict[object, object],
        streams: Sequence[object],
        ctm: Matrix = MATRIX_IDENTITY,
    ) -> None:
        # 重载返回指令流
        """Render the content streams.

        This method may be called recursively.
        """
        # log.debug(
        #     "render_contents: resources=%r, streams=%r, ctm=%r",
        #     resources,
        #     streams,
        #     ctm,
        # )
        self.init_resources(resources)
        self.init_state(ctm)
        return self.execute(list_value(streams))

    # Run PostScript commands
    # The Do_xxx method is the method for executing corresponding postscript instructions
    def execute(self, streams: Sequence[object]) -> None:
        # 重载返回指令流
        ops = ""
        try:
            parser = PDFContentParser(streams)
        except PSEOF:
            # empty page
            return
        while True:
            try:
                (_, obj) = parser.nextobject()
            except PSEOF:
                break
            if isinstance(obj, PSKeyword):
                name = keyword_name(obj)
                method = "do_%s" % name.replace("*", "_a").replace('"', "_w").replace(
                    "'",
                    "_q",
                )
                if hasattr(self, method):
                    func = getattr(self, method)
                    nargs = func.__code__.co_argcount - 1
                    if nargs:
                        args = self.pop(nargs)
                        # log.debug("exec: %s %r", name, args)
                        if len(args) == nargs:
                            func(*args)
                            if not (
                                name[0] == "T"
                                or name in ['"', "'", "EI", "MP", "DP", "BMC", "BDC"]
                            ):  # 过滤 T 系列文字指令，因为 EI 的参数是 obj 所以也需要过滤（只在少数文档中画横线时使用），过滤 marked 系列指令
                                p = " ".join(
                                    [
                                        (
                                            f"{x:f}"
                                            if isinstance(x, float)
                                            else str(x).replace("'", "")
                                        )
                                        for x in args
                                    ]
                                )
                                ops += f"{p} {name} "
                    else:
                        # log.debug("exec: %s", name)
                        targs = func()
                        if targs is None:
                            targs = []
                        if not (name[0] == "T" or name in ["BI", "ID", "EMC"]):
                            p = " ".join(
                                [
                                    (
                                        f"{x:f}"
                                        if isinstance(x, float)
                                        else str(x).replace("'", "")
                                    )
                                    for x in targs
                                ]
                            )
                            ops += f"{p} {name} "
                elif settings.STRICT:
                    error_msg = "Unknown operator: %r" % name
                    raise PDFInterpreterError(error_msg)
            else:
                self.push(obj)
        # print('REV DATA',ops)
        return ops
